/**
 * The file contains a simple Matrix [of doubles] class,
 * wrapping simple operations like addition, multiplication,
 * and maybe some LAPACK routines required in the project.
 *
 * Author: Konstantin Tretyakov
 * Created: 01.11.2004
 */

#ifndef __Matrix_h__
#define __Matrix_h__

#include <ostream>
#include <string>
#include <vector>
#include <istream>
using std::ostream;
using std::string;
using std::vector;
using std::istream;


// --------------------------------------------------------------- //

class Matrix {
private:
    int rows, cols;
    double* data; // Matrix values, in _column_major_ order
                  // (so it's easy to call LAPACK routines)
public:
    Matrix();
    Matrix(int rows, int cols, double d = 0);
    Matrix(const Matrix& m);
     // NB: This constructor accepts a double array in _row_major_ (C-type) order
    Matrix(int rows, int cols, const double* m);
    Matrix(int rows, int cols, istream& in);
    ~Matrix();
    
    int getRows() const;
    int getCols() const;
    
    // Get/set matrix element at given row/column (1-based index)
    double& operator()(int row, int col);
    double operator()(int row, int col) const;
    
    Matrix& operator=(const Matrix&m);
    bool operator==(const Matrix& m) const;
    bool operator!=(const Matrix& m) const;
    Matrix operator*(const Matrix& m) const;
    Matrix operator*(double d) const;
    Matrix operator/(double d) const;
    Matrix operator+(const Matrix& m) const;
    Matrix operator-(const Matrix& m) const;
    Matrix operator|(const Matrix& m) const; // Concatenate horizontally
    Matrix operator&(const Matrix& m) const; // Concatenate vertically
    Matrix transpose() const;
    Matrix getRowRange(int from, int to) const;
    Matrix getColRange(int from, int to) const;
    double trace() const;
    
    void debug() const;

    struct GenEigenvector;
        
    // Similar to SciLab-s "spec". Returns generalized vectors
    friend vector<Matrix::GenEigenvector*> Spec(Matrix a, Matrix b);
};


// -------------------------------------------------------------------------- //

struct Matrix::GenEigenvector {
    double alphar;
    double alphai;
    double beta;
    Matrix eigenvector;
    GenEigenvector(double alphar, double alphai, double beta, const Matrix& eigenvector):
        alphar(alphar), alphai(alphai), beta(beta), eigenvector(eigenvector) {}

};

    
// ---------- Functions ----------------------------------------------------- //

ostream& operator<<(ostream& out, const Matrix& m);
inline Matrix operator*(double d, const Matrix& m) { return m*d; };


// ---------- Inlines ------------------------------------------------------- //

inline double& Matrix::operator()(int row, int col) {
    if (row < 1 || row > rows || col < 1 || col > cols) throw string("Invalid matrix subscript");
    return data[rows * (col - 1) + row - 1];   
}

inline double Matrix::operator()(int row, int col) const {
    if (row < 1 || row > rows || col < 1 || col > cols) throw string("Invalid matrix subscript");
    return data[rows * (col - 1) + row - 1];
}    


inline int Matrix::getRows() const {
    return rows;
}


inline int Matrix::getCols() const {
    return cols;
}

inline bool Matrix::operator!=(const Matrix& m) const {
    return !(this->operator==(m));
}

#endif
