/**
 * Matrix class code.
 *
 * Author: Konstantin Tretyakov
 * Created: 01.11.2004
 */

#include "Matrix.h"
#include <cstring>
#include <string>
#include <iostream>
using std::endl;
using std::string;
using namespace std;


// -------------------------------------------------------------------------- //

// External functions from BLAS:
#include "blaswrap.h"
// Multiplication (calculates alpha*op(A)*op(B) + beta*C)
extern "C" int dgemm_(const char& transa, const char& transb, 
    const long& m, const long& n, const long& k, 
    const double& alpha, const double *a, const long& lda, 
	const double *b, const long& ldb, const double& beta, double *c__, 
	const long& ldc);

// Generalized eigenvalues routine
#include "dggev.h"


// -------------------------------------------------------------------------- //

Matrix::Matrix():
    rows(0), cols(0), data(NULL) { };


// -------------------------------------------------------------------------- //

Matrix::Matrix(int rows, int cols, double d): 
    rows(rows), cols(cols), data(NULL) {
    if (rows < 0 || cols < 0) throw string("Matrix dimensions must be non-negative"); 
    long datasize = rows*cols*sizeof(double);
    if (datasize > 0) {
        data = (double*)malloc(datasize);
        if (d == 0) memset(data, 0, datasize);
        else 
            for (int i = 0; i < rows*cols; i++) data[i] = d;
    }
}


// -------------------------------------------------------------------------- //

Matrix::Matrix(const Matrix& m){
    rows = m.rows;
    cols = m.cols;
    if (m.data == NULL) data = NULL;
    else {
        data = (double*)malloc(rows*cols*sizeof(double));
        memcpy(data, m.data, rows*cols*sizeof(double));
    }
}


// -------------------------------------------------------------------------- //

Matrix::Matrix(int rows, int cols, const double* m):
    rows(rows), cols(cols), data(NULL) {
    if (rows < 0 || cols < 0) throw string("Matrix dimensions must be non-negative");
    if (rows > 0 && cols > 0) {
        data = (double*)malloc(rows*cols*sizeof(double));
        for (int i = 0; i < rows; i++)
            for (int j = 0; j < cols; j++)
                data[rows*j + i] = m[cols*i + j];
    }
}


// -------------------------------------------------------------------------- //

Matrix::Matrix(int rows, int cols, istream& in):
    rows(rows), cols(cols), data(NULL) {
    if (rows < 0 || cols < 0) throw string("Matrix dimensions must be non-negative");
    if (rows > 0 && cols > 0) {
        data = (double*)malloc(rows*cols*sizeof(double));
        for (int i = 0; i < rows; i++)
            for (int j = 0; j < cols; j++)
                in >> data[rows*j + i];
    }
}


// -------------------------------------------------------------------------- //

Matrix::~Matrix(){
    if (data != NULL) free(data);
}


// -------------------------------------------------------------------------- //

Matrix& Matrix::operator=(const Matrix& m) {
    if (&m == this) return *this;
    rows = m.rows; cols = m.cols;
    if (m.data == NULL) {
        if (data != NULL) free(data);
        data = NULL;
    }
    else {
        data = (double*)realloc(data, m.rows*m.cols*sizeof(double));
        memcpy(data, m.data, rows*cols*sizeof(double));
    }
    return *this;
}


// -------------------------------------------------------------------------- //

bool Matrix::operator==(const Matrix& m) const {
    if (m.data == NULL && data == NULL) return true;
    if (m.rows != rows || m.cols != cols) return false;
    return (memcmp(data, m.data, rows*cols*sizeof(double)) == 0);
}


// -------------------------------------------------------------------------- //

Matrix Matrix::operator*(const Matrix& m) const {
    if (data == NULL || m.data == NULL) return Matrix();
    if (cols != m.rows) throw string("Matrix dimensions don't match");
    Matrix result(rows, m.cols);
    dgemm_('n', 'n', rows, m.cols, cols, 
                    1.0, data, rows, m.data, m.rows, 0.0, result.data, rows);
    return result;
}


// -------------------------------------------------------------------------- //

Matrix Matrix::operator*(double d) const {
    if (data == NULL) return Matrix();
    Matrix result(rows, cols);
    for (int i = 0; i < rows * cols; i++) result.data[i] = data[i]*d;
    return result;
}


// -------------------------------------------------------------------------- //

Matrix Matrix::operator/(double d) const {
    if (data == NULL) return Matrix();
    Matrix result(rows, cols);
    for (int i = 0; i < rows * cols; i++) result.data[i] = data[i]/d;
    return result;
}


// -------------------------------------------------------------------------- //

Matrix Matrix::operator+(const Matrix& m) const {
    if ((data == NULL && m.data != NULL) ||
        (data != NULL && m.data == NULL) ||
         rows != m.rows || cols != m.cols) throw string("Matrix dimensions don't match");
    if (data == NULL) return Matrix();
    
    Matrix result(rows, cols);
    for (int i = 0; i < rows*cols; i++)
        result.data[i] = data[i] + m.data[i];
    return result;
}


// -------------------------------------------------------------------------- //

Matrix Matrix::operator-(const Matrix& m) const {
    if ((data == NULL && m.data != NULL) ||
        (data != NULL && m.data == NULL) ||
         rows != m.rows || cols != m.cols) throw string("Matrix dimensions don't match");
    if (data == NULL) return Matrix();
    
    Matrix result(rows, cols);
    for (int i = 0; i < rows*cols; i++)
        result.data[i] = data[i] - m.data[i];
    return result;
}


// -------------------------------------------------------------------------- //

// Concatenate horizontally
Matrix Matrix::operator|(const Matrix& m) const {
    if (data == NULL) return m;
    if (m.data == NULL) return *this;
    if (rows != m.rows) 
        throw string("Matrix dimensions don't match");
    Matrix result(rows, cols + m.cols);
    memcpy(result.data, data, rows*cols*sizeof(double));
    memcpy(&result.data[rows*cols], m.data, m.rows * m.cols * sizeof(double));
    return result;
}


// -------------------------------------------------------------------------- //

// Concatenate vertically
Matrix Matrix::operator&(const Matrix& m) const {
    if (data == NULL) return m;
    if (m.data == NULL) return *this;
    if (cols != m.cols) 
        throw string("Matrix dimensions don't match");
    Matrix result(rows + m.rows, cols);
    for (int j = 1; j <= cols; j++) {
        for (int i = 1; i <= rows; i++)
            result(i, j) = (*this)(i, j);
        for (int i = 1; i <= m.rows; i++)
            result(i + rows, j) = m(i, j);
    } 
    return result;
}


// -------------------------------------------------------------------------- //

Matrix Matrix::transpose() const {
    if (data == NULL) return *this;
    if (rows == 1 || cols == 1) {
        Matrix result = *this;
        result.rows = cols;
        result.cols = rows;
        return result;
    }
    Matrix result(cols, rows);
    for (int i = 1; i <= rows; i++)
        for (int j = 1; j <= cols; j++)
            result(j, i) = (*this)(i, j);
    return result;
}


// -------------------------------------------------------------------------- //

double Matrix::trace() const {
    if (rows != cols) throw string("Trace is only defined for square matrices");
    if (data == NULL) return 0;
    
    double result = 0.0;
    for (int i = 1; i <= rows; i++) result += (*this)(i, i);
    return result;
}


// -------------------------------------------------------------------------- //

Matrix Matrix::getRowRange(int from, int to) const {
    if (data == NULL) throw string("Cannot get row range of an empty matrix");
    if (from < 1 || to > rows || from > to) throw string("Invalid matrix dimensions");
    Matrix result(to - from + 1, cols);
    for (int j = 1; j <= cols; j++)
        for (int i = from; i <= to; i++)
            result(i - from + 1, j) = (*this)(i, j);
    return result;
}


// -------------------------------------------------------------------------- //

Matrix Matrix::getColRange(int from, int to) const {
    if (data == NULL) throw string("Cannot get col range of an empty matrix");
    if (from < 1 || to > cols || from > to) throw string("Invalid matrix dimensions");
    Matrix result(rows, to - from + 1);
    for (int j = from; j <= to; j++)
        memcpy(&result.data[rows*(j - from)], &data[rows*(j - 1)], rows*sizeof(double));
    return result;
}


// -------------------------------------------------------------------------- //

void Matrix::debug() const {
    cout << "Data: " << (unsigned long)data << endl;
    cout << "Rows: " << rows << endl;
    cout << "Cols: " << cols << endl;
}


// -------------------------------------------------------------------------- //

vector<Matrix::GenEigenvector*> Spec(Matrix a, Matrix b) {
    double* alphar;
    double* alphai;
    double* beta;
    double* vr;
    double* work;
    double worksize;
    long worksizeint;
    long info;
    
    int n = a.rows;
    if (a.data == NULL || b.rows != n || a.cols != n || b.cols != n)
        throw string("DGGEV expecting two square matrices for A and B");
        
    // Determine required workspace size
    int res = dggev_('N', 'V', n, a.data, n, b.data, n, 
                        alphar, alphai, beta, NULL, 1, vr, n, &worksize, -1, &info);
    
    if (info != 0) throw string("DGGEV failed to determine workspace size");
     
    // Create required structures
    alphar = new double[n];
    alphai = new double[n];
    beta   = new double[n];
    vr     = new double[n*n];
    //memset(vr, 0, n*n*sizeof(double)); // Useful for debugging
    worksizeint = (long)worksize;
    work = new double[worksizeint];
    
    // Note: DGGEV overwrites a.data and b.data but we don't care cos we took parameters by
    // value.
    res = dggev_('N', 'V', n, a.data, n, b.data, n, 
                        alphar, alphai, beta, NULL, 1, vr, n, work, worksizeint, &info);

    if (info != 0) throw string("DGGEV failed");
    
    // Fill the result structure
    vector<Matrix::GenEigenvector*> result;
    for (int i = 0; i < n; i++) {
        result.push_back(new Matrix::GenEigenvector(alphar[i], alphai[i], beta[i], 
                                Matrix(a.rows, 1, &vr[i*a.rows])));
    }
    
    delete[] work;
    delete[] vr;
    delete[] beta;
    delete[] alphai;
    delete[] alphar;
    
    return result;
}


// -------------------------------------------------------------------------- //

ostream& operator<<(ostream& out, const Matrix& m) {
    for (int i = 1; i <= m.getRows(); i++) {
        for (int j = 1; j <= m.getCols(); j++) {
            out << m(i, j);
            if (j != m.getCols()) out << " ";
        }
        out << endl;
    }
    return out;
}


