// ------------------------------------------------------------------------------//
// Kernel-based Regularized Canonical Correlation Analysis for Cross-Language Retrieval.
// (C) 2004  Konstantin Tretyakov, Jelena Zaitseva
//


// ------------------------------------------------------------------------------//
// Kernel regularized CCA
//
// Given kernels K_x, K_y corresponding to data samples X and Y finds k pairs of 
// directions w_x(i) and w_y(i) [i=1..k] such that X-s projection on w_x is maximally
// correlated with Y-s projection on w_y, in particular
//   (w_x, w_y) maximise (X*w_x)'*(Y*w_y)
//        subject to (X*w_x)'*(X*w_x) = (Y*w_y)'*(Y*w_y) = 1
//
// Parameters: 
//              K_x, K_y - Kernel matrices for X and Y-s (K_x = X*X', K_y = Y*Y')
//              k        - How many direction pairs to find
//              reg      - regularization parameter
//
// Returns:     W_x      - has vectors w_x as columns
//              W_y      - has vectors w_y as columns
//
function [W_x, W_y] = kernel_rcca(K_x, K_y, k, reg)
	// The solutions (w_x w_y) to kRCCA are the generalized eigenvalues of 
	//    A*[w_x; w_y] = lambda*B*[w_x; w_y]
	// where
	z=zeros(K_x);
	A = [    z       K_x*K_y ;...
    		 	K_y*K_x      z    ];
  // and
  B = [ (1 - reg)*K_x^2 + reg*K_x                      z    ;...
                 z                (1 - reg)*K_y^2 + reg*K_y];

  // To avoid numerical mistakes make B symmetric and positive definite           
	B =(B + B')/2;
  B = B + 1e-10*eye(B);

  // Solve for generalized eigenvalues (takes some time)
  [al, be, Z] = spec(A, B);
  
  // Find k largest eigenvalues (they must all be real, we seek for positive only cos
  // for each positive eigenvalue there is a similar negative one).
  // Eigenvalues are in al./be where be != 0.
  idx_nonzero_be = find(be);
	eigs = real(al(idx_nonzero_be)./be(idx_nonzero_be));
	[sorted_eigs, ranks] = sort(eigs);
	
	// Take k best ranks
	W = Z(:,ranks(1:k));
	W_x = W(1:size(K_x, 'r'), :);
	W_y = W((size(K_x,'r')+1):(2*size(K_x,'r')), :);
	
  // Normalize w_i so that w_i'*K_i^2*w_i = 1
  K_xx = K_x*K_x; K_yy = K_y*K_y;
  for i=1:size(W,'c')
    w_i = W_x(:,i);
  	 W_x(:,i) = w_i/sqrt(w_i' * K_xx * w_i);
  	 w_i = W_y(:,i);
  	 W_y(:,i) = w_i/sqrt(w_i' * K_yy * w_i);
  end


// ------------------------------------------------------------------------------//
// Trains a model for cross-language retrieval
//
// Parameters: 
//    K_x, K_y - X and Y kernels to use for training
//    k        - dimension of the canonical correlation vectors
//    reg      - regularization parameter
// Returns:
//    a structure that can be used in clr_evaluate
//
function clr_model = clr_train(K_x, K_y, k, reg)
	// Use kernel centering and normalization
	K_x_n = normalize_kernel(K_x);
	K_x_nc = center_kernel(K_x_n);
	K_y_n = normalize_kernel(K_y);
	K_y_nc = center_kernel(K_y_n);

	// Perform kRCCA on training data (takes time)
	printf("Performing kRCCA, please wait...");
	[W_x, W_y] = kernel_rcca(K_x_nc, K_y_nc, k, reg);
	printf(" Done\n");

	// Save results in a tlist and return
	clr_model = tlist(["clr_model", "K_x", "K_y", "K_x_n", "K_y_n", "W_x", "W_y"],...
	                                 K_x,   K_y,   K_x_n,   K_y_n,   W_x,   W_y);



// ------------------------------------------------------------------------------//
// Given a trained CLR model returns a semantic vector for a test sample x of language X
// 
// Parameters:
//   clr_model - model returned by clr_train
//   k         - vector of kernel evaluations for the test sample x
//   k_xx      - kernel value K(x, x).
// Returns:
//   a semantic vector corresponding to x
//
function v = clr_evaluate_x(clr_model, k, k_xx)
	// Normalize/center the sample
	k_x = center_test_sample(normalize_test_sample(k, clr_model("K_x"), k_xx), clr_model("K_x_n"));
	
	// Project k_x onto canonical components
	// Projection = < w, x > = < \sum w_i*x_i, x > = \sum w_i < x_i, x > = w'*k
	v = clr_model("W_x")'*k_x;


// ------------------------------------------------------------------------------//
// Same as above, but for language Y
// 
function v = clr_evaluate_y(clr_model, k, k_yy)
	// Normalize/center the sample
	k_y = center_test_sample(normalize_test_sample(k, clr_model("K_y"), k_yy), clr_model("K_y_n"));
	
	// Project k_x onto canonical components
	// Projection = < w, x > = < \sum w_i*x_i, x > = \sum w_i < x_i, x > = w'*k
	v = clr_model("W_y")'*k_y;
	


// ------------------------------------------------------------------------------//
// Evaluates clr model precision.
// 
// Parameters:
//   clr_model   - clr_model returned by clr_train
//   k_x_samples - matrix containing k_x test samples as columns
//   k_y_samples - matrix containing corresponding k_y test samples as columns
//                 (size(k_x) = size(k_y))
//   k_xx        - vector [K(x_test_i, x_test_i)] of length n.
//   k_yy        - vector [K(y_test_i, y_test_i)].
//   C           - integer. For each sample in x we find C semantically closest samples in y
//                 and consider it a match if the right sample is among them.
// Returns:
//   precision as a real between 0 and 1

function precision = clr_test(clr_model, k_x_samples, k_y_samples, k_xx, k_yy, C)
	// Find semantic vectors for all articles
	n = size(k_x_samples, 'c');
	Proj_x = [];
	Proj_y = [];
	for i=1:n
		proj_x = clr_evaluate_x(clr_model, k_x_samples(:,i), k_xx(i));
		proj_y = clr_evaluate_y(clr_model, k_y_samples(:,i), k_yy(i));
		Proj_x = [Proj_x proj_x];
		Proj_y = [Proj_y proj_y];
	end

	// For each article in X find whether the correct Y translation is within first C best matches.
	matches = 0;
	for i=1:n
		d = [];
		for j=1:n
			v = Proj_x(:,i) - Proj_y(:,j);
			d = [d v'*v];
		end;
		[tmp, bestmatches] = sort(-abs(d));
		matches = matches + length(find(bestmatches(1:C) == i));
	end

	precision = matches/n

