/*
 * Cross-language text retrieval main model code file
 *
 * Author: Konstantin Tretyakov
 * Created: 21.11.2004
 */

#include "CrosslangModel.h"
#include "Kernels.h"
#include <iostream>
#include <fstream>
#include <string>
#include <cassert>
using namespace std;

// --------------------------- CrosslangOptions ----------------------------- //

const string CrosslangOptions::USAGE_MESSAGE = 
    "Usage: \n"
    "  crosslang train --model=<model.clm> --traindata=<traindata.txt>\n"
    "                  --kernel=<kernelType> --k=<k> --reg=<reg>\n"
    "  crosslang query --model=<model.clm> --data=<data.txt> --query=<query>\n"
    "                  --matchcount=<m>\n"
    "  crosslang interactive --model=<model.clm> --data=<data.txt> --matchcount=<m>\n"
    "\n"
    " - Training data file contains 2n sentences, one per line, first n are in\n"
    "   language X and last n -- in language Y.\n"
    " - Data file contains language Y sentences to search from, one per line.\n"
    " - Query is a sentence in language X.\n"
    " - Kernel is specified as an integer kernelType. kernelType = 0 means BoW\n"
    "   kernel, kernelType = 1..6 means K-mer kernel\n"
    " - reg is the regularization parameter (default is 0.01) and k - number of \n"
    "   canonical components (default is 20)\n"
    " - matchcount specifies the number of matches to search for in the data\n"
    "   (default is 3)\n";


// -------------------------------------------------------------------------- //

CrosslangOptions::CrosslangOptions(int argc, char* argv[]) {
    
    // Set default values
    modelFile = "model.clm";
    trainFile = "traindata.txt";
    dataFile = "data.txt";
    kernelType = 0;
    k = 20;
    reg = 0.01;
    matchCount = 3;
    
    if (argc < 2) throw string("No action specified.");
    
    // Read action
    action = argv[1];
    if (action != "train" && action != "query" && action != "interactive") throw string("Invalid parameters.");
    
    // Read options
    for(int i = 2; i < argc; i++) {
        string curOption = argv[i];
        if (curOption.substr(0, 2) != "--") throw string("Invalid parameters.");
        int sepPos = curOption.find("=");
        if (sepPos < 0) throw string("Invalid parameters.");
        string optionName = curOption.substr(2, sepPos - 2);
        string optionValue = curOption.substr(sepPos + 1);
        if (optionName == "model") modelFile = optionValue;
        else if (optionName == "traindata") trainFile = optionValue;
        else if (optionName == "data") dataFile = optionValue;
        else if (optionName == "query") query = optionValue;
        else if (optionName == "kernel") {
           if (action != "train") throw string("No need to specify kernel type.");
           sscanf(optionValue.c_str(), "%d", &kernelType);
        }
        else if (optionName == "reg") {
           if (action != "train") throw string("No need to specify reg.");
           sscanf(optionValue.c_str(), "%f", &reg);
        }
        else if (optionName == "k") {
           if (action != "train") throw string("No need to specify k.");
           sscanf(optionValue.c_str(), "%d", &k);
        }
        else if (optionName == "matchcount") sscanf(optionValue.c_str(), "%d", &matchCount);
        else throw (string("Invalid parameter: ") + optionName);
    }
    
    // Sanity checks
    if (action == "query" && query == "") throw string("Query not specified.");
    if (kernelType > 6 || kernelType < 0) throw string("Wrong kernel type specified.");
    if (matchCount <= 0) throw string("Match count must be positive");
    
    #ifdef DEBUG
    cout << "Action: " << action << endl;
    cout << "Model file: " << modelFile << endl;
    cout << "Training data file: " << trainFile << endl;
    cout << "Data file: " << dataFile << endl;
    cout << "Kernel type: " << kernelType << endl;
    cout << "K: " << k << endl;
    cout << "Reg: " << reg << endl;
    cout << "Query: " << query << endl;
    #endif
}


// ------------------------------ CrosslangModel----------------------------- //

// Kind of a "main" method (not very clear OO, but it's short at least)
CrosslangModel::CrosslangModel(const CrosslangOptions& opts): 
    options(opts), trained(false) {
    // Either we train the model based on training data, or
    // Evaluate it on given data
    if (options.action == "train") {
        cout << "Loading training samples..." << endl;
        vector<string> instances = LoadStrings(options.trainFile.c_str());
        n = instances.size();
        if (n == 0 || n % 2 != 0) 
            throw string("Expecting even number of strings for training data.");
        n = n / 2;
        
        // Split training instances to X and Y
        copy(instances.begin(), instances.begin() + n, 
                insert_iterator<vector<string> >(trainX, trainX.begin()));
        copy(instances.begin() + n, instances.end(),
                insert_iterator<vector<string> >(trainY, trainY.begin()));
        cout << "Loaded " << n << " samples for each language." << endl;
        assert(trainX.size() == trainY.size());
        
        // Train model
        if (options.k > n) options.k = n - 5;
        if (options.k < 1) options.k = 1;
        train(trainX, trainY, options.k, options.reg);
    
        cout << "Saving model..." << endl;
        save(options.modelFile);
    }
    else if (options.action == "query") {
        load(options.modelFile);
        data = LoadStrings(options.dataFile.c_str());
        prepareData(data);
               
        Matrix q = evaluateX(options.query);
        vector<int> closest = findClosestSemanticVectors(q, options.matchCount);
        
        for (int i = 0; i < closest.size(); i++)
           cout << data[closest[i] - 1] << endl;
    }
    else if (options.action == "interactive") {
        cout << "Loading model..." << endl;
        load(options.modelFile);
        cout << "Loading data..." << endl;
        data = LoadStrings(options.dataFile.c_str());
        cout << "Preparing data..." << endl;
        prepareData(data);
        
        cout << "Enter query. Enter \\q to quit." << endl;
        string line;
        while (true) {
            cout << "> ";
            getline(cin, line);
            if (line == "\\q") break;
            Matrix q = evaluateX(line);
            vector<int> closest = findClosestSemanticVectors(q, options.matchCount);
            cout << options.matchCount << " best matches: " << endl << endl;
            for (int i = 0; i < closest.size(); i++)
                cout << closest[i] << ": " << data[closest[i] - 1] << endl << endl;
        }
        cout << "Bye!" << endl;
    }
}


// -------------------------------------------------------------------------- //

// Train model
void CrosslangModel::train(const vector<string>& trainX, const vector<string>& trainY, 
                           int k, double reg) {
    this->trainX = trainX;
    this->trainY = trainY;

    // Calculate kernels
    cout << "Calculating kernel matrices..." << endl;
    kernelX = EvaluateStringKernel(options.kernelType, trainX);
    kernelY = EvaluateStringKernel(options.kernelType, trainY);
    kernelXn = NormalizeKernel(kernelX);
    kernelYn = NormalizeKernel(kernelY);
    kernelXnc = CenterKernel(kernelXn);
    kernelYnc = CenterKernel(kernelYn);
    
    // Run kRCCA
    cout << "Running kRCCA..." << endl;
    Matrix w = KernelRCCA(kernelXnc, kernelYnc, k, reg);
    wX = w.getColRange(1, k);
    wY = w.getColRange(k+1, 2*k);

    cout << "Done." << endl;
    trained = true;
}


// -------------------------------------------------------------------------- //

// Precalculate dataSemanticVectors
void CrosslangModel::prepareData(const vector<string>& data) {
    if (!trained) throw string("Cannot call prepareData on untrained model");
    this->data = data;
    
    // Evaluate kernel for data items and Y-language training items
    //cout << "Evaluating kernel on data to be queried..." << endl;
    vector<string> trainAndData = trainY;
    trainAndData.insert(trainAndData.end(), data.begin(), data.end());
    Matrix dataKernel = EvaluateStringKernel(options.kernelType, trainAndData);
    dataKernel = dataKernel.getColRange(trainY.size() + 1, trainAndData.size());
    dataKernel = dataKernel.getRowRange(1, trainY.size());
    
    // Check..
    int t = rand() % data.size();
    assert(dataKernel.getColRange(t+1, t+1) == EvaluateStringKernel(options.kernelType, data[t], trainY));
    
    // Now dataKernel(:,i) is a vector of kernel evaluations for data item i.
    //cout << "Finding data semantic vectors..." << endl;
    dataSemanticVectors = vector<Matrix>();
    
    for (int i = 1; i <= data.size(); i++) {
       dataSemanticVectors.push_back(evaluateY(dataKernel.getColRange(i, i), data[i - 1]));
    }   
}


// -------------------------------------------------------------------------- //

// Save model to file
void CrosslangModel::save(const string& filename) const {
    // Save model to file
    ofstream out(filename.c_str());
    if (!out) throw (string("Cannot open file ") + filename + " for writing");
    
    out << "CrosslangModelV0.1" << endl;
    out << n << endl;
    for (vector<string>::const_iterator i = trainX.begin(); i != trainX.end(); i++)
        out << *i << endl;
    for (vector<string>::const_iterator i = trainY.begin(); i != trainY.end(); i++)
        out << *i << endl;
    out << options.k << " " << options.kernelType << endl;
    out << wX << wY << endl;
    
    out << kernelX << kernelXn << kernelXnc << endl;
    out << kernelY << kernelYn << kernelYnc << endl;
    out.close();
}


// -------------------------------------------------------------------------- //

// Load model from file
void CrosslangModel::load(const string& filename) {
    // Load model from file
    ifstream in(filename.c_str());
    if (!in) throw (string("Cannot open file ") + filename + " for reading");
    
    string line;
    getline(in, line);
    if (line != "CrosslangModelV0.1") throw string("Invalid file format");
    in >> n;
    getline(in, line);
    assert(line == "");
    
    // Load strings used for training
    trainX = LoadStrings(in, n);
    trainY = LoadStrings(in, n);
    assert(trainX.size() == n && trainY.size() == n);
    
    // Load kernels and canonical vectors
    in >> options.k >> options.kernelType;
    wX = Matrix(n, options.k, in);
    wY = Matrix(n, options.k, in);
    kernelX = Matrix(n, n, in);
    kernelXn = Matrix(n, n, in);
    kernelXnc = Matrix(n, n, in);
    kernelY = Matrix(n, n, in);
    kernelYn = Matrix(n, n, in);
    kernelYnc = Matrix(n, n, in);
    in.close();
    trained = true;
}


// -------------------------------------------------------------------------- //

// Finds a semantic vector for phrase in lang. X
Matrix CrosslangModel::evaluateX(const string& s) {
    if (!trained) throw string("Cannot call evaluate on untrained model");
    Matrix kx = EvaluateStringKernel(options.kernelType, s, trainX);
    double kxx = EvaluateStringKernel(options.kernelType, s, s);
    kx = CenterTestSample(NormalizeTestSample(kx, kernelX, kxx), kernelXn);   
    return (kx.transpose()*wX).transpose();
}


// -------------------------------------------------------------------------- //

// Finds a semantic vector for phrase in lang. Y
Matrix CrosslangModel::evaluateY(const string& s) {
    if (!trained) throw string("Cannot call evaluate on untrained model");
    Matrix ky = EvaluateStringKernel(options.kernelType, s, trainY);
    double kyy = EvaluateStringKernel(options.kernelType, s, s);
    ky = CenterTestSample(NormalizeTestSample(ky, kernelY, kyy), kernelYn);    
    return (ky.transpose()*wY).transpose();
}


// -------------------------------------------------------------------------- //

// Finds a semantic vector for a phrase in lang. Y
// Note that here we give not a string but already a vector of 
// kernel evaluations.
Matrix CrosslangModel::evaluateY(const Matrix& k, const string& s) {
    if (!trained) throw string("Cannot call evaluate on untrained model");
    double kyy = EvaluateStringKernel(options.kernelType, s, s);
    Matrix ky = CenterTestSample(NormalizeTestSample(k, kernelY, kyy), kernelYn);
    return (ky.transpose()*wY).transpose();
}
    

// -------------------------------------------------------------------------- //
    
// Finds k closest matches to given vector among the semantic vectors
vector<int> CrosslangModel::findClosestSemanticVectors(const Matrix& q, int k) {
    if (!trained) throw string("Cannot use untrained model");
    vector<pair<double, int> > dists;
    for (int i = 1; i < dataSemanticVectors.size(); i++) {
        Matrix m = (q - dataSemanticVectors[i-1]);
        double dist = (m.transpose()*m)(1,1);
        pair<double, int> pd(dist, i);
        dists.push_back(pd);
    };
    
    partial_sort(dists.begin(), dists.begin() + k, dists.end());
    
    vector<int> result;
    for (int i = 0; i < k; i++) {
       result.push_back(dists[i].second);
    }
    return result;
}


