#include "Kernels.h"

#include <set>
#include <iostream>
#include <fstream>
using namespace std;

//----------------------------------------------------------------------------------------//
// Checks used in the next function
bool isWhitespace(unsigned char c){
    return (c == ' ' || c < 32 || c == '\t' || c == '\n');
}

bool isDigit(unsigned char c) {
    return (c <= '9' && c >= '0');
}

bool isPunctuation(unsigned char c) {
    return (c >= 33 && c <= 47) || (c >= 58 && c <= 64) || (c >=91 && c <=96)
           || (c >= 123 && c <= 127);
}

bool isLetter(unsigned char c) {
    return !(isWhitespace(c) || isPunctuation(c) || isDigit(c));
}

char toLower(unsigned char c) {
    if (c <= 'Z' && c >= 'A') return c - 'A' + 'a';
    else return c;
}

/**
 * Strips punctuation, and converts to lowercase
 */
string PreprocessString(const string& s) {
    string result;
    for (int i = 0; i < s.length(); i++) {
        if (isLetter(s[i])) result += toLower(s[i]);
        else if (isWhitespace(s[i])) {
            result += ' ';
            while (i < s.length() && !isLetter(s[i])) i++;
            i--;
        }
   }
   return result;
}


//----------------------------------------------------------------------------------------//
/**
 * Helper function for kernel computation. Returns an inner product of two maps
 */
template <typename T, typename U>
U mapInnerProduct(const map<T, U>& a, const map<T, U>& b) {
    U result = 0;
    typename map<T, U>::const_iterator itA = a.begin();
    typename map<T, U>::const_iterator itB = b.begin();
    while (itA != a.end() && itB != b.end()) {
        T keyA = itA->first;
        T keyB = itB->first;

        if (keyA < keyB) itA++;
        else if (keyB < keyA) itB++;
        else {
            result += itA->second * itB->second;
            itA++; itB++;
        }
    }
    return result;
}

// ---------------------------------------------------------------------------------------//

/**
 * O(l*log(l))
 */
double BoWKernel::operator()(const string& s1, const string& s2) const {
    string a = PreprocessString(s1);
    string b = PreprocessString(s2);
    
   // Count all words in the first string (put them all in a map)
	int curpos = 0;
	int predpos = 0;
	map<string, long> bowA;
	do{	
	    curpos = a.find(" ", curpos);
	    string curBoW = a.substr(predpos, curpos - predpos);
	    if (curBoW != "") bowA[curBoW]++;
	    predpos = ++curpos;
	} while (curpos > 0);
	    
    // Do the same for the second string
	curpos = 0;
	predpos = 0;
	map<string, long> bowB;
	do{	
	    curpos = b.find(" ", curpos);
	    string curBoW = b.substr(predpos, curpos - predpos);
	    if (curBoW != "") bowB[curBoW]++;
	    predpos = ++curpos;
	} while (curpos > 0);

    // Now traverse both maps simultaneously counting similar words
    return (double) mapInnerProduct(bowA, bowB);
}


// ---------------------------------------------------------------------------------------//

/**
 * Calculates the kernel matrix by first calculating for each word the number of
 * its occurences in each phrase, and then for each word updating the matrix.
 * Therefore the complexity is O(l*log(l)*n + C*n*n) where C depends on how often an average
 * occurs. 
 * Reason for speedup is the fact that we split phrases into words only once, and
 * then for each word we only consider the phrases where it occurs.
 */
Matrix BoWKernel::getKernelMatrix(const vector<string>& strings) const {
    if (strings.size() == 0) throw string("At least one instance expected");
    vector<string> instances(strings.size());
    for (int i = 0; i < strings.size(); i++) {
        instances[i] = PreprocessString(strings[i]);
    }
    
    // Fill a map: Word -> (Map: phraseIndex -> number of occurences)
    map<string, map<int, int> > phrasesContainingWord;
    
    int idx = 0;
    for (vector<string>::const_iterator i = instances.begin(); i != instances.end(); i++) {
        
        // Split to words
        int curpos = 0;
        int predpos = 0;
    	do{	
    	    curpos = i->find(" ", curpos);
    	    string curWord = i->substr(predpos, curpos - predpos);
    	    if (curWord != "") phrasesContainingWord[curWord][idx]++;
    	    predpos = ++curpos;
    	} while (curpos > 0);
	
    	idx++;
   	}
   	
   	// Now fill the result matrix
   	Matrix kernel(instances.size(), instances.size());
   	for (map<string, map<int, int> >::iterator i = phrasesContainingWord.begin();
   	     i != phrasesContainingWord.end(); i++) {
   	
   	   map<int, int> wordMap = i->second;
   	   for (map<int, int>::iterator j = wordMap.begin(); j != wordMap.end(); j++)
           for (map<int, int>::iterator k = wordMap.begin(); k != wordMap.end(); k++)
               kernel(j->first + 1, k->first + 1) += j->second * k->second;
    }
    
    return kernel;
}

// ---------------------------------------------------------------------------------------//


KMerKernel::KMerKernel(int k): k(k) {
    ;
}


// ---------------------------------------------------------------------------------------//

/**
 * Direct computation of K-Mer kernel for two given strings
 * in O(l*log(l)) steps where lengths of the strings are O(l).
 */
double KMerKernel::operator()(const string& s1, const string& s2) const {
    string a = PreprocessString(s1);
    string b = PreprocessString(s2);
    
    if (a.length() < k || b.length() < k) return 0;
    
    // Count all k-mers in the first string (put them all in a map)
    map<string, long> kMersA;
    for (int i = 0; i <= a.length() - k; i++) {
        string curKmer = a.substr(i, k);
        kMersA[curKmer]++;
    }
    
    // Do the same for the second string
    map<string, long> kMersB;
    for (int i = 0; i <= b.length() - k; i++) {
        string curKmer = b.substr(i, k);
        kMersB[curKmer]++;
    }
    
    // Now traverse both maps simultaneously counting similar k-mers
    return (double) mapInnerProduct(kMersA, kMersB);
}


// ---------------------------------------------------------------------------------------//

/**
 * Calculates the kernel matrix in O(A^k * n + l*n) where A is the alphabet size
 * Is good when k is small and there many instances.
 */
Matrix KMerKernel::getKernelMatrix(const vector<string>& strings) const {
    if (strings.size() == 0) throw string("At least one instance expected");
    vector<string> instances(strings.size());
    for (int i = 0; i < strings.size(); i++) instances[i] = PreprocessString(strings[i]);
    
    // Determine target alphabet
    set<char> charset;
    for (vector<string>::const_iterator i = instances.begin(); i != instances.end(); i++){
        for (int j = 0; j < i->length(); j++) charset.insert((*i)[j]);
    }
    
    // Convert set to char*
    char* alphabet = new char[charset.size()];
    
    int k = 0;
    for (set<char>::iterator i = charset.begin(); 
         i != charset.end(); i++)
        alphabet[k++] = *i;

    // Calculate matrix by traversing the trie
    Matrix result(instances.size(), instances.size());
    map<int, vector<int> > empty;
    traverseTrie(instances, alphabet, charset.size(), 0, empty, result);
    delete[] alphabet;
    
    return result;
}


// ---------------------------------------------------------------------------------------//

/**
 * Trie traversal to calculate the whole kernel matrix at once.
 * Parameters: 
 *    instances - strings for which the matrix is being calculated
 *    alphabet  - alphabet over which the trie is constructed (e.g. set of all chars in strings)
 *    alphabetSize - size of the alphabet
 *    curDepth  - current depth in the trie. 0 means trie 
 *    positions - for each string that contains substring corresponding to current trie node
 *                contains a vector of positions at which that string ends
 *    k         - kernel matrix to be updated.
 */
void KMerKernel::traverseTrie(const vector<string>& instances, const char* alphabet, int alphabetSize, int curDepth,  const map<int, vector<int> >& positions, Matrix& kernel) const {
    
    if (curDepth == k) {
        // This is a leaf node, update kernel matrix
        for (map<int, vector<int> >::const_iterator i = positions.begin(); i != positions.end(); i++) {
            for (map<int, vector<int> >::const_iterator j = positions.begin(); j != i; j++) {
                kernel(i->first + 1, j->first + 1) += i->second.size() * j->second.size();
                kernel(j->first + 1, i->first + 1) += i->second.size() * j->second.size();
            }
            kernel(i->first + 1, i->first + 1) += i->second.size() * i->second.size();  
        }
    }
    else if (curDepth == 0) {
        // This is an initial node, calculate positions in a bit different way
        for (int c = 0; c < alphabetSize; c++) {
            map<int, vector<int> > newPositions;
            // For each string find the positions of char alphabet[c]
            for (int i = 0; i < instances.size(); i++){
                int firstpos = instances[i].find(alphabet[c]);
                while (firstpos >= 0) {
                    newPositions[i].push_back(firstpos);
                    firstpos = instances[i].find(alphabet[c], firstpos + 1);
                } 
            }
            if (newPositions.size() > 0)
                traverseTrie(instances, alphabet, alphabetSize, curDepth + 1, newPositions, kernel);
        }
    }
    else {
        
        // Traverse subleafs of the trie
        for (int c = 0; c < alphabetSize; c++){
            map<int, vector<int> > newPositions;
            
            // For each string find the positions where a substring (corresponding to current
            // trie leaf) ends.
            for (map<int, vector<int> >::const_iterator i = positions.begin(); 
                i != positions.end(); i++){
                int instIndex = i->first;
                vector<int> instPositions = i->second;
                for (vector<int>::const_iterator pos = instPositions.begin();
                     pos != instPositions.end(); pos++) {
                         
                    // pos indicates the position where substring of the current level ended
                    // check if the next position contains the right character
                    if (*pos + 1 < instances[instIndex].length() && 
                        (instances[instIndex])[*pos + 1] == alphabet[c])
                        newPositions[instIndex].push_back(*pos + 1);
                }
            }
            
            // Traverse subleaf c
            if (newPositions.size() > 0) 
                traverseTrie(instances, alphabet, alphabetSize, curDepth + 1, newPositions, kernel);
             
        } // for c in alphabet
    }
        
}


// --------------------------------------------------------------------- //

/**
 * Loads a vector of strings from file.
 */
vector<string> LoadStrings(const char* filename) {
        // Read strings from file
        ifstream in(filename);
        if (!in) throw (string("Could not open file ") + filename);
        vector<string> result = LoadStrings(in, -1);
        in.close();
        return result;
}


// --------------------------------------------------------------------- //

/**
 * Loads a vector of strings from stream. 
 * If n >= 0, n strings are loaded, otherwise all strings until eof
 */
vector<string> LoadStrings(istream& in, int n) {
    string line;
    vector<string> result;
    while(in && n-- != 0) {
        getline(in, line);
        if (line != "") result.push_back(line);
    }
    
    return result;
}


// --------------------------------------------------------------------- //

/**
 * Calculates kernel matrix from a vector of strings
 * instances - vector of strings, kernelType - kernel type (0 - bow, k - k-mer)
 */
Matrix EvaluateStringKernel(int kernelType, const vector<string>& instances) {
    if (kernelType == 0) {
        BoWKernel kernel;
        return kernel.getKernelMatrix(instances);
    }
    else {
        KMerKernel kernel(kernelType);
        return kernel.getKernelMatrix(instances);
    }
}

// --------------------------------------------------------------------- //

Matrix EvaluateStringKernel(int kernelType, const string& s, const vector<string>& v) {
    int n = v.size();
    Matrix kx(n, 1);
    
    if (kernelType == 0) {
        BoWKernel bow;
        for (int i = 0; i < n; i++)
            kx(i+1, 1) = bow(s, v[i]);
    } 
    else {
        KMerKernel kmer(kernelType);
        for (int i = 0; i < n; i++)
            kx(i+1, 1) = kmer(s, v[i]);
    }
    
    return kx;
}


// --------------------------------------------------------------------- //

double EvaluateStringKernel(int kernelType, const string& s, const string& t) {
    if (kernelType == 0) {
        BoWKernel bow;
        return bow(s, t);
    } 
    else {
        KMerKernel kmer(kernelType);
        return kmer(s, t);
    }
}


// --------------------------------------------------------------------- //

Matrix NormalizeKernel(const Matrix& kernel) {
    Matrix result = kernel;
    
    // Divide every element (i, j) by sqrt(K(i,i))*sqrt(K(j, j));
    for (int i = 1; i <= result.getRows(); i++)
        for (int j = 1; j <= result.getCols(); j++)
            result(i, j) /= sqrt(kernel(i, i) * kernel(j, j));
            
    return result;
}


// --------------------------------------------------------------------- //

Matrix CenterKernel(const Matrix& kernel) {
    int n = kernel.getRows();
    Matrix ones(n, n, 1.0);
    return kernel - ones*kernel/n - kernel*ones/n + ones*kernel*ones/n/n;
}

// --------------------------------------------------------------------- //

// kXX = unnormalized kernel evaluated at (x, x)
Matrix NormalizeTestSample(const Matrix& k, const Matrix& K, double kXX) {
    Matrix kn = k;
    for (int i = 1; i <= kn.getRows(); i++) {
        kn(i, 1) /= sqrt(K(i, i) * kXX);
    }
    return kn;
}

// --------------------------------------------------------------------- //

Matrix CenterTestSample(const Matrix& k, const Matrix& K) {
    int n = k.getRows();
    Matrix j(n, 1, 1.0);
    Matrix J(n, n, 1.0);
    return k - J*k/n - K*j/n + J*K*j/n/n;
}

// -------------------------------------------------------------------- //
// Comparator for eigenvalues.
class eigcompare {
public:
    bool operator()(Matrix::GenEigenvector* a, Matrix::GenEigenvector* b) {
            if (b->beta == 0 && a->beta == 0) return (a->alphar > b->alphar);
            else if (b->beta == 0) return true;
            else if (a->beta == 0) return false;
            else return (a->alphar/a->beta > b->alphar/b->beta);
    }
};

/**
 * Performs kernel rCCA on (kX, kY).
 * k specifies the number of canonical components to return.
 * Returns the matrix with 2k columns. First k columns are X canonical components,
 * second - Y-s.
 */
Matrix KernelRCCA(const Matrix& K_x, const Matrix& K_y, int k, double reg) {
    int n = K_x.getRows();
    if (k < 1 || k > n) throw string("Invalid parameter k value");
    if (K_x.getCols() != n || K_y.getRows() != n || K_y.getCols() != n)
        throw string("Matrix dimensions don't match");

    Matrix zeros(n, n);
    Matrix K_xx = K_x * K_x;
    Matrix K_yy = K_y * K_y;
    Matrix A = (   zeros   | (K_x*K_y) ) &
               ( (K_y*K_x) |   zeros   );
               
    Matrix B = ( ((1 - reg)*K_xx + reg*K_x) |           zeros            ) &
               (         zeros              | ((1 - reg)*K_yy + reg*K_y) );
    
    Matrix ridge(2*n, 2*n);
    for (int i = 1; i <= 2*n; i++) ridge(i, i) = 1e-10;
    // Make B symmetric and positive definite
    B = (B + B.transpose())/2 + ridge;
    
    vector<Matrix::GenEigenvector*> eigs = Spec(A, B);
    
    // Get k best eigenvectors
    partial_sort(eigs.begin(), eigs.begin() + k, eigs.end(), eigcompare());
    
    // Return result (each eigenvalue consists of corresponding
    //   X canonical component and Y canonical component. Split them and return)
    // We also normalize all vectors w_i so that w_i'*K_i^2*w_i = 1
    Matrix result;

    for (int i = 0; i < k; i++) {
        Matrix wx = eigs[i]->eigenvector.getRowRange(1,n);
        double norm = sqrt((wx.transpose()*K_xx*wx)(1,1));
        wx = wx / norm;
        result = result | wx;
    }

    for (int i = 0; i < k; i++) {
        Matrix wy = eigs[i]->eigenvector.getRowRange(n+1, 2*n);
        double norm = sqrt((wy.transpose()*K_yy*wy)(1,1));
        wy = wy / norm;
        result = result | wy;
    }

    // Free eigs
    for (int i = 0; i < eigs.size(); i++) delete eigs[i];
    
    return result;
}


