/** 
 * Kernel-related classes and functions.
 *
 * Author: Konstantin Tretyakov
 * Created: 22.11.2004
 */

#ifndef __Kernels_h__
#define __Kernels_h__

#include "Matrix.h"
#include <string>
#include <map>
#include <vector>
#include <iostream>
using std::string;
using std::map;
using std::vector;
using std::istream;


// ---------------------------- Kernel classes -------------------------------//

/**
 * A kernel is a class with an () operator that takes 2 parameters
 * and returns a double.
 *
 * K-mer kernel is inner product in a feature space indexed by
 * all k-mers.
 */
class KMerKernel {
public:
    const int k;
    
    KMerKernel(int k);
    double operator()(const string& s1, const string& s2) const;
    Matrix getKernelMatrix(const vector<string>& strings) const;
    
private:
    void traverseTrie(const vector<string>& instances, const char* alphabet, int alphabetSize, int curDepth,  const map<int, vector<int> >& positions, Matrix& k) const; 
};


// -------------------------------------------------------------------------- //

/**
 * Bag of words kernel
 */
class BoWKernel {
public:
    double operator()(const string& s1, const string& s2) const;
    Matrix getKernelMatrix(const vector<string>& strings) const;
};



// ------------------------------ Subroutines --------------------------------//

/**
 * Given a kernel K and a set of instances calculates the kernel matrix
 * by applying the kernel to all pairs of instances.
 * This method is slower than custom methods, but may be used as a reference,
 * and is applicable in any case.
 * O(K*n^2) where K is kernel computation complexity.
 */
template <typename Kernel, typename Data>
Matrix GetKernelMatrix(const Kernel& K, const vector<Data>& instances) {
    if (instances.size() == 0) throw string("At least one instance expected");
    Matrix result(instances.size(), instances.size());
    
    for (int i = 1; i <= instances.size(); i++)
        for (int j = 1; j <= i; j++)
            result(i, j) = result(j, i) = K(instances[i - 1], instances[j - 1]);
    return result;
}


// -------------------------------------------------------------------------- //

/**
 * Loads a vector of strings from file.
 */
vector<string> LoadStrings(const char* filename);

/**
 * Loads a vector of strings from stream. 
 * If n >= 0, n strings are loaded, otherwise all strings until eof [or long overflow :)]
 */
vector<string> LoadStrings(istream& in, int n);

/**
 * Calculates kernel matrix from a vector of strings
 * instances - vector of strings, kernelType - kernel type (0 - bow, k - k-mer)
 */
Matrix EvaluateStringKernel(int kernelType, const vector<string>& instances);
Matrix EvaluateStringKernel(int kernelType, const string& s, const vector<string>& v);
double EvaluateStringKernel(int kernelType, const string& s, const string& t);

/**
 * Strips punctuation, and converts to lowercase
 */
string PreprocessString(const string& s);

Matrix NormalizeKernel(const Matrix& kernel);

Matrix CenterKernel(const Matrix& kernel);

Matrix NormalizeTestSample(const Matrix& k, const Matrix& kernel, double kernelXX);

Matrix CenterTestSample(const Matrix& k, const Matrix& kernel);

/**
 * Performs kernel rCCA on (kernelX, kernelY).
 * k specifies the number of canonical components to return.
 * Returns the matrix with 2k columns. First k columns are X canonical components,
 * second - Y-s.
 */
Matrix KernelRCCA(const Matrix& kernelX, const Matrix& kernelY, int k, double reg);


#endif

