#include "Kernels.h"
#include "Matrix.h"
#include <iostream>
#include <cassert>
#include <ctime>
#ifndef CLK_TCK
 #define CLK_TCK CLOCKS_PER_SEC
#endif
using namespace std;


int main() {
    try {
        srand(10000);

        // Test k-mer kernel
        KMerKernel kmer3(3);
        assert(kmer3("abcde","") == 0);
        assert(kmer3("","") == 0);
        assert(kmer3("abcabc", "abcabc") == 4 + 1 + 1);
        assert(kmer3("aaaaaaaaaaaaaaaaa", "aaaaaaaaaaaaaaaa") == 15*14);
        assert(kmer3("ababababababababa", "abababa") == 8*3 + 7*2);
        
        // Test bow kernel
        BoWKernel bow;
        assert(bow("abc", "cde") == 0);
        assert(bow("", "") == 0);
        assert(bow("abc", "") == 0);
        assert(bow("abc", "abc") == 1);
        assert(bow("to be or not to be", "be this good or not good") == 2*1 + 1*1 + 1*1);
        assert(bow(" a  b   c  a  ", " b   a b") == 2*1 + 2*1);
        
        // Test utility routines
        assert(PreprocessString("") == "");
        assert(PreprocessString(" \t \t\n \t") == " ");
        assert(PreprocessString("Linux 2.6 is  good!\n") == "linux is good ");
        
        // Performance test for KMerKernel
        vector<string> instances;
        int n = 10;
        char chars[] = { 'A', 'T', 'O', 'M', ' ' };
        for (int i = 0; i < n; i++){
            string newS;
            for (int j = 0; j < 1000; j++) newS += chars[rand() % 5];
            //cout << newS << endl;
            instances.push_back(newS);
        }

        cout << "Calculating 3mer matrix for " << n << " strings, direct method: ";
        cout.flush();
        clock_t start = clock();
        Matrix k1 = GetKernelMatrix(kmer3, instances);
        cout << ((double)(clock() - start))/CLK_TCK << endl;


        cout << "Using trie: ";
        cout.flush();
        start = clock();
        Matrix k2 = kmer3.getKernelMatrix(instances);
        cout << ((double)(clock() - start))/CLK_TCK << endl;

        //cout << k1 << endl;
        assert(k1 == k2);


        // Performance test for BoW
        cout << "Calculating BoW matrix for " << n << " strings, direct method: ";
        cout.flush();
        start = clock();
        k1 = GetKernelMatrix(bow, instances);
        cout << ((double)(clock() - start))/CLK_TCK << endl;


        cout << "By words: ";
        cout.flush();
        start = clock();
        k2 = bow.getKernelMatrix(instances);
        cout << ((double)(clock() - start))/CLK_TCK << endl;

        assert(k1 == k2);
        
        // Test kernel/instance normalization/centering
        Matrix kn = NormalizeKernel(k1);
        assert(abs(kn.trace() - n) < 0.000001);
        Matrix inst = k1.getColRange(1,1);
        Matrix instN = NormalizeTestSample(inst, k1, k1(1,1));
        assert(instN == kn.getColRange(1,1));
        assert(abs(instN(1,1) - 1.0) < 0.0000001);
        
        Matrix kc = CenterKernel(k1);
        Matrix instC = CenterTestSample(inst, k1);
        assert(instC == kc.getColRange(1,1));
        
        cout << "Test successful" << endl;
    }
    catch (string s) {
        cout << "Exception: " << s << endl;
    }
    return 0;
}

