// ------------------------------------------------------------------------------//
// Kernel Adatron Algorithm for SVM Binary Classification
// (C) 2004  Konstantin Tretyakov, Jelena Zaitseva
//

// Requires kernel_methods.sci

// ------------------------------------------------------------------------------//
// The Kernel Adatron Algorithm.
//
// Input:
//              K	  - The kernel matrix for training samples
//	             y   - Vector of correct classifications of training samples
//                    (1 denotes correct class, -1 denotes incorrect class)
//              C   - Regularization parameter
//              eta - Step size (1.0 is good for normalized kernels)
// Returns:     
//              alpha, b - Trained SVM parameters.
//                    Note that alphas are already multiplied by y-s.
//
function [alphas, b] = kernel_adatron(K, y, C, eta)
// We search for alpha, b that classify samples correctly
	n = size(K, 'r'); // n = number of training samples
	j = ones(n, 1);
	alphas = zeros(n, 1);
	b = 0;

	for i=1:(10*n)
		
		// z is the prediction of the training set labels for current alpha,b
		z = K*(y.*alphas) + b;
		
		// Kinda gradient vector
		g = j - (y.*z);
		betas = min(max(alphas + eta * g, zeros(g)), C * ones(g));
		[elem, index] = max(abs(alphas - betas));
		alphas(index) = betas(index);
   
		// Update b
		b = b + y' * alphas/4;
	end
	
	alphas = y.*alphas


// ------------------------------------------------------------------------------//
// Two simple wrappers for kernel_adatron. svm_train trains on given data, and returns a model,
// svm_evaluate takes a vector of kernel evaluations for some training samples, and
// returns the response of a trained svm given by the svm parameter.
// Several samples are given as columns of k, result is a row vector then.
//
function svm = svm_train(K, y, C)
	// Use kernel normalization (the adatron algorithm likes that)
	K_n = normalize_kernel(K);
	[alphas, b] = kernel_adatron(K_n, y, C, 1.0);
	svm = tlist(["svm", "alphas", "b", "K"], alphas, b, K);

// k_xx = K(x, x), required to normalize the training sample
function y = svm_evaluate(svm, k, k_xx)
	k_n = normalize_test_sample(k, svm("K"), k_xx);
	y = k_n' * svm("alphas") + svm("b");


// ------------------------------------------------------------------------------//
// Simple Kernel Adatron Test
//
function kernel_adatron_test()
	
	// Data
	X = [0 -1; -1 -1; -1 0; 1 0; 0 1; 1 1];
	
	// Labels
	y = [1, 1, 1, -1, -1, -1]';
	n = length(y);
	
	// Kernel
	K = X*X';
	
	// Train
	[alphas, b] = kernel_adatron(K, y, 10, 1);

  for i=1:100
    for j=1:100
      p_X(i)=0.02*i-1;
      p_Y(j)=0.02*j-1;
      p_Z(i,j)=sign(svm_evaluate(X*[p_X(i); -p_Y(j)], alphas, b));
    end
  end

  r = [-1.5, -1.5, 1.5, 1.5];
	Matplot1(p_Z' + 3, r);
	plot2d(X(find(y == 1),1),X(find(y == 1),2),-5,rect=r);
	plot2d(X(find(y == -1),1),X(find(y == -1),2),-4,rect=r);


