// ------------------------------------------------------------------------------//
// Collection of generic kernel-related functions.
// (C) 2004  Konstantin Tretyakov, Jelena Zaitseva
//

// ------------------------------------------------------------------------------//
// Kernel centering
// 
// Suppose X is the data matrix. Then K = X*X'.
// Centered data matrix is X_c = (X - J*X/n) where J = ones(K)
// Therefore centered kernel matrix is
// K_c = (X - J*X/n)*(X' - X'*J/n) = X*X' - J*X*X'/n - X*X'*J/n + J*X*X'*J/n^2 =
//     = K - J*K/n - K*J/n + J*K*J/n^2
//
function K_c = center_kernel(K)
	n = size(K, 'r');
	J = ones(K);
	K_c = K - J*K/n - K*J/n + J*K*J/n^2;


// ------------------------------------------------------------------------------//
// Kernel normalization
//
// Normalized data matrix is
//  X_n = diag(x_1'*x_1, x_2'*x_2, ..., x_n'*x_n)^(-1/2) * X
//
// Therefore
//  K_n = diag(K_11, K_22, ..., K_nn)^(-1/2)* X_n * X_n' * diag(K_11...)^(-1/2)' =
//      = K_d * K * K_d   where K_d = diag(K_11, ..., K_nn)^(-1/2)
//
function K_n = normalize_kernel(K)
	n = size(K, 'r');
	K_d = diag(diag(K))^(-1/2);
	K_n = K_d * K * K_d;


// ------------------------------------------------------------------------------//
// Test sample centering corresponding to kernel centering
// 
// Given a vector of kernel evaluations k = K(x, x_i) for test sample x
// return a centered version k_c = K(centered(x), centered(x_i)) corresponding
// to centering of kernel by center_kernel procedure
//
// Let x be the sample to center, k = X*x - matrix of kernel evaluations for this sample 
// on uncentered data, K = X*X' - the (uncentered) kernel matrix with respect to which 
// we center, x_i_c = centered training samples, X_c data matrix of centered samples,
// j = column vector of n ones, J = j'*j.
// Then
//   x_c = x - X'*j/n
//   k_c = X_c * x_c = (X - J*X/n)*(x - X'*j/n) = X*x - J*X*x/n - X*X'*j/n + J*X*X'*j/n^2 =
//       = k - J*k/n - K*j/n + J*K*j/n^2
//
function k_c = center_test_sample(k, K)
	n = size(K,'r');
	j = ones(n, 1);
	J = j*j';
	k_c = k - J*k/n - K*j/n + J*K*j/n^2


// ------------------------------------------------------------------------------//
// Test sample normalization corresponding to kernel normalization
// 
// Given a vector of kernel evaluations k = K(x, x_i) for test sample x
// return a normalized version k_n = K(normalized(x), normalized(x_i)) corresponding
// to normalization of kernel by normalize_kernel procedure
//
// If x is the sample to normalize, k is the corresponding dual vector, then
//   x_n = x/sqrt(x'*x)
//   k_n = X_n*x_n = diag(x_1'*x_1, x_2'*x_2, ..., x_n'*x_n)^(-1/2) * X * x/sqrt(x'*x) =
//       = diag(K_11, .. ,K_nn)^(-1/2) * k/sqrt(K(x, x))
//
// Parameters: 
//   k - sample to normalize
//   K - (unnormalized) kernel matrix
//   K_xx - evaluation of the unnormalized kernel on the training sample x
//
function k_n = normalize_test_sample(k, K, K_xx)
	n = size(K, 'r');
	K_d = diag(diag(K))^(-1/2);
	k_n = K_d * k/sqrt(K_xx);


// ------------------------------------------------------------------------------//
// Split given vector v into training/test set randomly
//  0 <= p <= 1 specifies the proportion of training set
//
function [train, test] = random_split(v, p)
	[tmp, randperm] = sort(rand(v));
	len = length(randperm);
	len_train = round(len*p);
	train = v(randperm(1:len_train));
	test  = v(randperm((len_train+1):$));


// ------------------------------------------------------------------------------//
// Plot kernel (more-or-less like matlab's imagesc)
//
function plot_kernel(K)
	xbasc();
	cm = 0:0.01:1.0;
	cm = cm^(1/1.9); // Correct for 1.9 gamma so the picture will be brighter
	xset("colormap", [0*cm' 0*cm' cm']); // Imitate bluish Matlab output
	k_tmp = abs(K);
	k_tmp = 100*k_tmp/max(k_tmp); // Rescale values to 0..100
	Matplot(k_tmp, strf="020");



