// ------------------------------------------------------------------------------//
// Convenience Routines for Multiclass Classification
// (C) 2004  Konstantin Tretyakov, Jelena Zaitseva
//

// Requires kernel_adatron.sci

// ------------------------------------------------------------------------------//
// Trains a One-versus-All multiclass classifier.
//
// Input:
//        K       - kernel matrix of training samples
//        labels  - class labels for training samples
//        C       - regularization parameter
//                  (NB: It is not right to use a single regularization parameter, but
//                       that's the requirement of the assignment)
// Output:
//        ova     - structure that may be used in ova_evaluate, ova_test
//
function ova = ova_train(K, labels, C)

	classes = union(labels, []);
	n_classes = length(classes);
	
	// For each class find a classifier that distinguishes it from others
	// Note that it is not right to use a single regularization parameter for all classifiers,
	// the right way is to use a separate C for each binary classifier
	svms = list();
	for i = 1:n_classes
		y = zeros(labels) -  1;
		y(find(labels == classes(i))) = 1;
		svms(i) = svm_train(K, y, C);
	end

	ova = tlist(["ova", "svms", "classes"], svms, classes);


// ------------------------------------------------------------------------------//
// Returns the prediction of a trained OvA classifier.
//
// Input: 
//         ova  - classifier as returned by ova_train
//         k    - vector of kernel evaluations for the sample
//         k_xx - K(x,x) for the sample
//
// Output: 
//         class label for the sample as predicted by the classifier
//
function c = ova_evaluate(ova, k, k_xx)
	evals = [];
	for i = 1:length(ova("classes"))
		evals = [evals, svm_evaluate(ova("svms")(i), k, k_xx)];
	end
	[m, which] = max(evals);
	c = ova("classes")(which);


// ------------------------------------------------------------------------------//
// Given a trained OvA classifier and a set of test samples returns classifier precision
//
// Input: 
//         ova    - classifier as returned by ova_train
//         k      - test samples as columns (k(:,i) = K(x_train(:), x_test(i))
//         k_xx   - vector of K(x,x)-s (k_xx(i) = K(x_test(i), x_test(i))
//         labels - correct labels for the samples
//
// Output: 
//         Classifier precision
//
function p = ova_test(ova, k, k_xx, labels)

	// Evaluate classifier on all test samples
	evals = [];
	n = size(k, "c");
	for i = 1:n
		c = ova_evaluate(ova, k(:,i), k_xx(i));
		evals = [evals; c];
	end
	
	p = length(find(labels == evals)) / n;


// ------------------------------------------------------------------------------//
// Trains a All-versus-All multiclass classifier. Usage similar to ova_train.
//
function ava = ava_train(K, labels, C)

	classes = union(labels, []);
	n_classes = length(classes);
	
	// For each class pair find a classifier that distinguishes the two classes
	// Note that it is not right to use a single regularization parameter for all classifiers,
	// the right way is to use a separate C for each binary classifier
	svm_lists = list(); // svm_lists(i) is a list of svms that treat class i as positive.
	for i = 1:n_classes
		svm_lists(i) = list();
		idx_i = find(labels == classes(i));
		for j = 1:(i-1)
			idx_j = find(labels == classes(j));
			y = [ones(idx_i), -1 * ones(idx_j)]';
			svm_lists(i)(j) = svm_train(K([idx_i, idx_j], [idx_i, idx_j]), y, C);
		end
	end

	ava = tlist(["ava", "svm_lists", "classes", "labels"], svm_lists, classes, labels);


// ------------------------------------------------------------------------------//
// Returns the prediction of a trained AvA classifier. Usage similar to ova_evaluate.
//
function c = ava_evaluate(ava, k, k_xx)
	// Evaluate all classifiers on the training samples. Assign the class getting the
	// largest sum of scores.
	evals = [];
	n_classes = length(ava("classes"));
	for i = 1:n_classes
		idx_i = find(ava("labels") == ava("classes")(i));
		for j = 1:(i-1)
			idx_j = find(ava("labels") == ava("classes")(j));
			k_ij = k([idx_i, idx_j]);		
			evals(i, j) = svm_evaluate(ava("svm_lists")(i)(j), k_ij, k_xx);
			evals(j, i) = -evals(i, j);
		end
	end
	evals = sum(evals, 'c');
	[m, cl] = max(evals);
	c = ava("classes")(cl);


// ------------------------------------------------------------------------------//
// Given a trained OvA classifier and a set of test samples returns classifier precision
//
function p = ava_test(ava, k, k_xx, labels)

	// Evaluate classifier on all test samples
	evals = [];
	n = size(k, "c");
	for i = 1:n
		c = ava_evaluate(ava, k(:,i), k_xx(i));
		evals = [evals; c];
	end
	
	p = length(find(labels == evals)) / n;

