// ------------------------------------------------------------------------------//
// Multiclass Text Classification with Kernel Adatron SVM learning
// (C) 2004  Konstantin Tretyakov, Jelena Zaitseva
//

clear 
K_3mer = read("../data/K_3mer.dat", 262, 262);
K_bow  = read("../data/K_bow.dat", 262, 262);

rand('seed', 300);
getf("kernel_methods.sci"); // Load auxiliary subroutines
getf("kernel_adatron.sci"); 
getf("kernel_multiclass.sci");

// Kernel to use
// We normalize it but do not center it, because the
// Kernel adatron algorithm assumes that diagonal elements are not too
// different from the inverse of the step size (1 in our case)
K = K_bow;

// Extract subkernel for English
n = size(K, 'r') / 2;
K = K(1:n, 1:n);

// Mark chapters for each paragraph
clear para_chapter;
para_chapter(1:5) = 1;
para_chapter(6:23) = 2;
para_chapter(24:32) = 3;
para_chapter(33:53) = 4;
para_chapter(54:69) = 5;
para_chapter(70:80) = 6;
para_chapter(81:92) = 7;
para_chapter(93:97) = 8;
para_chapter(98:105) = 9;
para_chapter(106:118) = 10;
para_chapter(119:126) = 11;
para_chapter(127:129) = 12;
para_chapter(130:131) = 13;

// We are only interested in chapters 2-7, 9-10
chapters=[2:7, 9:10];
idx = [];
for c = chapters
	idx = [idx, find(para_chapter == c)];
end

K = K(idx, idx);
para_chapter = para_chapter(idx);

// Stratified split into train (80%) and test (20%) data
idx_train = [];
idx_test  = [];
for c = chapters	
	[idx_train_c, idx_test_c] = random_split(find(para_chapter == c), 0.8);
	idx_train = [idx_train, idx_train_c];
	idx_test  = [idx_test,  idx_test_c];
end


// Determine best C using stratified leave-one-out cross validation
C_range = [1e-4 1e-3 1e-2 0.1 1 10 100];
ava_cv_precisions = [];
ova_cv_precisions = [];
ava_test_precisions = [];
ova_test_precisions = [];

for idx_C=1:length(C_range)
	C = C_range(idx_C);
	printf("Iterating for C=%f\n", C);
	ava_cv_precisions(idx_C) = 0;
	ova_cv_precisions(idx_C) = 0;
	ava_test_precisions(idx_C) = 0;
	ova_test_precisions(idx_C) = 0;
	
	for t=1:length(idx_train)  // on each iteration idx_train(t) is left out for testing
		idx_trn = idx_train([1:(t-1), (t+1):length(idx_train)]);
		idx_val = idx_train(t);
		
		ava = ava_train(K(idx_trn, idx_trn), para_chapter(idx_trn), C);
		p_ava = ava_test(ava, K(idx_trn, idx_val), diag(K(idx_val, idx_val)), para_chapter(idx_val));
		ava_cv_precisions(idx_C) = ava_cv_precisions(idx_C) + p_ava/length(idx_train);
		
		ova = ova_train(K(idx_trn, idx_trn), para_chapter(idx_trn), C);
		p_ova = ova_test(ova, K(idx_trn, idx_val), diag(K(idx_val, idx_val)), para_chapter(idx_val)); 
		ova_cv_precisions(idx_C) = ova_cv_precisions(idx_C) + p_ova/length(idx_train);
		
		printf("Done t: %d of %d, p(ava): %f, p(ova): %f\n", t, length(idx_train), p_ava, p_ova);
	end
	
	// Check test precisions for this C
	ava = ava_train(K(idx_train, idx_train), para_chapter(idx_train), C);
	p_ava = ava_test(ava, K(idx_train, idx_test), diag(K(idx_test, idx_test)), para_chapter(idx_test));
	ava_test_precisions(idx_C) = p_ava;
		
	ova = ova_train(K(idx_train, idx_train), para_chapter(idx_train), C);
	p_ova = ova_test(ova, K(idx_train, idx_test), diag(K(idx_test, idx_test)), para_chapter(idx_test)); 
	ova_test_precisions(idx_C) = p_ova;
	
	printf("Cross-validation precision for C=%f: p_ava=%f, p_ova=%f\n", C, ava_cv_precisions(idx_C), ova_cv_precisions(idx_C));
	printf("Test precision             for C=%f: p_ava=%f, p_ova=%f\n", C, p_ava, p_ova);
end


// Save results
save("multiclass_classification.dat", ava_cv_precisions, ova_cv_precisions, ava_test_precisions, ova_test_precisions, C_range);

// -------------------- EOF -------------------- //


